import os

from ament_index_python.packages import get_package_share_directory
from pathlib import Path
from launch import LaunchDescription
from launch.actions import ExecuteProcess, DeclareLaunchArgument, OpaqueFunction, ExecuteProcess, IncludeLaunchDescription, TimerAction
from launch_ros.actions import Node
from launch_param_builder import load_xacro
from launch.launch_description_sources import PythonLaunchDescriptionSource
from launch.conditions import LaunchConfigurationEquals


ld = LaunchDescription()
launch_file_dir = os.path.join(get_package_share_directory('atlantis_gazebo'), 'launch')
atlantis_scenarios_dir = get_package_share_directory('atlantis_scenarios')
controller_dir = get_package_share_directory('pmb2_controller_configuration')


def set_up_simulation(context):

    x = '2.0'
    y = '-2.0'
    z = '0.5'
    yaw = '0.0'
    robot = 'pmb2'
    world = os.path.join(atlantis_scenarios_dir, 'scenarios','apartment', 'apartment.world')
    
    default_controllers_cmd = IncludeLaunchDescription(
        PythonLaunchDescriptionSource(os.path.join(controller_dir,'launch', 'default_controllers.launch.py')),
        launch_arguments={'is_public_sim': 'True'}.items())
        
    remappings = [('/tf', 'tf'),
                    ('/tf_static', 'tf_static')]

    xacro_file_path = Path(os.path.join(
            get_package_share_directory('pmb2_description'),
            'robots',
            'pmb2.urdf.xacro',
        ))
    
    xacro_input_args = {
        'laser_model': 'sick-571',
        'add_on_module': 'no-add-on',
        'use_sim_time': 'True',
        'is_public_sim': 'True',

    }
    robot_description = load_xacro(xacro_file_path, xacro_input_args)

    start_gazebo_server_cmd = ExecuteProcess(
        cmd=['gzserver', '--verbose', '-s', 'libgazebo_ros_init.so', '-s', 'libgazebo_ros_factory.so', world],
        cwd=[launch_file_dir], output='screen')

    start_gazebo_client_cmd = ExecuteProcess(
        cmd=['gzclient', '--verbose'],
        cwd=[launch_file_dir], output='screen')
    
    state_publisher_cmd = Node(
        name=  'robot_state_publisher',
        package='robot_state_publisher',
        executable='robot_state_publisher',
        output='screen',
        parameters=[
            {'use_sim_time': True,
            'robot_description': robot_description}
        ],
        remappings=remappings)
    
    start_gazebo_spawner_cmd = Node(
        package='gazebo_ros',
        executable='spawn_entity.py',
        arguments=[
            '-topic', 'robot_description',
            '-entity', robot,
            '-x', x,
            '-y', y,
            '-z',  z,
            '-Y', yaw,
        ],
        output='screen',
    )

    teleop_twist_keyboard_cmd = Node(
        condition=LaunchConfigurationEquals('lab', ''),
            package='teleop_twist_keyboard',
            executable='teleop_twist_keyboard',
            name='teleop_keyboard',
            output='screen',
            remappings=[('/cmd_vel', '/mobile_base_controller/cmd_vel_unstamped')],
             prefix='xterm -e'
        )

    start_lab1 = ExecuteProcess(
        condition=LaunchConfigurationEquals('lab', 'fuzzy_controller_lab'),
        cmd=['python3', 'src/navigo/navigo/fuzzy_controller_lab.py'],
        cwd=["./"], output='screen')
    
    start_lab2 = ExecuteProcess(
        condition=LaunchConfigurationEquals('lab', 'planning_lab'),
        cmd=['python3', 'src/navigo/navigo/planning_lab.py'],
        cwd=["./"], output='screen')
    
    start_lab3 = ExecuteProcess(
        condition=LaunchConfigurationEquals('lab', 'knowledge_graph_lab'),
        cmd=['python3', 'src/navigo/navigo/knowledge_graph_lab.py'],
        cwd=["./"], output='screen')

  
    ld.add_action(start_gazebo_server_cmd)
    ld.add_action(start_gazebo_client_cmd)
    ld.add_action(start_gazebo_spawner_cmd)
    ld.add_action(default_controllers_cmd)
    ld.add_action(state_publisher_cmd)
    ld.add_action(teleop_twist_keyboard_cmd)
    ld.add_action(TimerAction(period=5.0, actions=[start_lab1, start_lab2, start_lab3]))


def generate_launch_description():
    # Default dirs

    # Create the launch configuration variables
    declare_lab_cmd = DeclareLaunchArgument(
        'lab',
        default_value='',
        description='Name of the lab to use')
    

    ld.add_action(OpaqueFunction(function=set_up_simulation))

    # Declare the launch options
    ld.add_action(declare_lab_cmd)
    

    return ld