"""
============== AIoD school on AI and Robotics 2024 ==============
Base code: dummy controller, performs the action decision parts of the main loop
To be customized in an incremental way for the different labs

(c) 2024 Alessandro Saffiotti
"""
from fcontrol import Behavior
from fcontrol import ramp_up, ramp_down, triangle, trapezoid
from fcontrol import global_to_local, local_to_global
from math import atan2, degrees, sqrt, pi

import robot_gwy


class GoToTarget (Behavior):
    """
    Navigate to a target object at (x,y,th,radius), where 'x,y' is
    the object's center in global frame, and 'radius' is its size
    The behavior stops at about 'radius' mt from the object's center
    The object's orientation th is ignored
    It assumes that there are no obstacles on the way
    """
    def __init__(self, target = (0.0, 0.0, 0.0, 0.0)):
        self.tpoint = [target[0], target[1], 0.0]   # target center point in global coordinates
        self.tlocal = [0, 0, 0]                     # target center point in robot's coordinates
        print("Target: " , target)
        self.radius = target[3]                     # distance to stop from target's centerpoint
        super().__init__()
    
    def update_state(self, state):
        """
        Set the local state variables phi and rho
        Use the 'mypose' estimate, passed as part of the global state
        """

        # 'phi' and 'rho' are radial coordinates to target
        mypose = state['mypose']
        global_to_local(self.tpoint, mypose, self.tlocal)
        xt = self.tlocal[0]
        yt = self.tlocal[1]
        self.state['phi'] = degrees(atan2(yt, xt))
        self.state['rho'] = sqrt(xt * xt + yt * yt)

    def setup(self):
        """
        Ppredicates, linguistic variables and rules of our 'go to' fuzzy behavior
        """
        robot_radius = 0.4
        stop_dist = self.radius + robot_radius

        self.fpreds = {
            # Fuzzy predicates used in the LHS of rules
            'TargetLeft'  : (ramp_up(0.0, 1.0), 'phi'),
            'TargetRight' : (ramp_down(-1.0, 0.0), 'phi'),
            'TargetAhead' : (triangle(-1.0, 0.0, 1.0), 'phi'),
            'TargetHere'  : (ramp_down(stop_dist, stop_dist+1.0), 'rho'),
        }

        self.flvars = {
            # Definition of the fuzzy linguistic variables, used in the rules' RHS
            'Move' : ({'Fast':0.5, 'Slow':0.1, 'None':0, 'Back':-0.1}, 'Vlin'),
            'Turn' : ({'Left':30, 'MLeft':10, 'None':0, 'MRight':-10, 'Right':-30}, 'Vrot')
        }

        self.frules = {
            'ToLeft'  : ("TargetLeft AND NOT(TargetHere)", 'Turn', 'Left'),
            'ToRight' : ("TargetRight AND NOT(TargetHere)", 'Turn', 'Right'),
            'Straight' : ("TargetAhead", 'Turn', 'None'),
            'Far'     : ("TargetAhead AND NOT(TargetHere)", 'Move', 'Fast'),
            'Stop'    : ("TargetHere OR NOT(TargetAhead)", 'Move', 'None')
        }

        # The degree of achievement is given by the predicate 'TargetHere'
        self.fgoal = "TargetHere"

        # Finally, initialize the fuzzy predicats and fuzzy output sets
        self.init_flsets()
        self.init_fpreds()


class Wander (Behavior):
    def __init__(self, target = (0.0, 0.0)):
        super().__init__()
    
    def update_state(self, state):
        """
        Set the local state variables for clearance (left, right, front, overall)
        Uses the 'sdata' sonar ranges, passed as part of the global state
        """
        ranges = [s[1] for s in state['sdata']]
        self.state['lft_clr'] = min(ranges[1], ranges[2])
        self.state['rgt_clr'] = min(ranges[10], ranges[11])
        self.state['frt_clr'] = ranges[0]
        self.state['all_clr'] = min(ranges)

    def setup(self):
        self.fpreds = {
            # Definition of the fuzzy predicates, used in the rules' LHS
            'ObstacleLeft'  : (ramp_down(0.2, 1.2), 'lft_clr'),
            'ObstacleRight' : (ramp_down(0.2, 1.2), 'rgt_clr'),
            'ObstacleAhead' : (ramp_down(0.5, 1.5), 'frt_clr'),
            # Fuzzy predicate do decide between avoid and go
            'Danger'        : (ramp_down(0.2, 2.0), 'all_clr')
        }

        self.flvars = {
            # Definition of the fuzzy linguistic variables, used in the rules' RHS
            'Move' : ({'Fast':0.7, 'Slow':0.1, 'None':0, 'Back':-0.1}, 'Vlin'),
            'Turn' : ({'Left':30, 'MLeft':10, 'None':0, 'MRight':-10, 'Right':-30}, 'Vrot')
        }

        self.frules = {
            # Lastly, definition of the actual fuzzy rules
            'GoLeft'   : ("ObstacleRight AND NOT(ObstacleLeft)", 'Turn', 'Right'),
            'GoRight'  : ("ObstacleLeft AND NOT(ObstacleRight)", 'Turn', 'Left'),
            'Tunnel'   : ("ObstacleRight AND ObstacleLeft", 'Turn', 'None'),
            'NoFront'  : ("NOT(ObstacleAhead)", 'Turn', 'None'),
            'Front'    : ("ObstacleAhead", 'Move', 'Back'),
            'Slow'     : ("Danger", 'Move', 'Slow'),
            'Fast'     : ("NOT(Danger)", 'Move', 'Fast')
        }

        # The degree of achievement is set to false: this behavior never ends
        self.fgoal = "False"

        # Finally, initialize the fuzzy predicats and fuzzy output sets
        self.init_flsets()
        self.init_fpreds()
            

class GoTo (Behavior):
    """
    Navigate to a target object at (x,y,th,radius), where 'x,y' is
    the object's center in global frame, and 'radius' is its size
    The behavior stops at about 'radius' mt from the object's center
    The object's orientation th is ignored
    Sonars are used for obstacle avoidance, using a modular approach
    """
    def __init__(self, target = (0.0, 0.0, 0.0, 0.0)):
        self.tpoint = [target[0], target[1], 0.0]   # target center point in global coordinates
        self.tlocal = [0, 0, 0]                     # target center point in robot's coordinates
        self.radius = target[3]                     # distance to stop from target's centerpoint
        print("Instantiated GoTo behavior with target:" , target)
        super().__init__()
    
    def update_state(self, state):
        """
        Hint: use a combination of the local variables in GoToTarget and in Wander
        """
        print("! Behavior GoTo not yet implemented")

    def setup(self):
        """
        Hint: use a combination of GoToTarget and Wander
        use the fuzzy predicate 'Danger' to decide between the two rulesets
        """
        robot_radius = 0.4
        stop_dist = self.radius + robot_radius

        self.fpreds = {
        }

        self.flvars = {
            # You may change these fuzzy linguistic variables as you see fit
            'Move' : ({'Fast':0.7, 'Slow':0.2, 'None':0, 'Back':-0.1}, 'Vlin'),
            'Turn' : ({'Left':30, 'MLeft':10, 'None':0, 'MRight':-10, 'Right':-30}, 'Vrot')
        }

        self.frules = {
        }

        # Set the degree of achievement
        self.fgoal = "True"

        # Finally, initialize the fuzzy predicats and fuzzy output sets
        self.init_flsets()
        self.init_fpreds()
            

class Cross (Behavior):
    def __init__(self, door = (0.0, 0.0, 0.0, 0.0), state = None):
        self.dpoint = [door[0], door[1], door[2]]    # door center point in global coordinates
        self.dlocal = [0, 0, 0]                      # door center point in robot's coordinates
        self.radius = door[3]                        # door radius
        print("Instantiated Cross behavior with door:", door)
        super().__init__()

    def update_state(self, state):
        print("! Behavior Cross not yet implemented")

    def setup(self):
        self.fpreds = {
        }

        self.flvars = {
            # You may change these fuzzy linguistic variables as you see fit
            'Move' : ({'Fast':0.4, 'Slow':0.2, 'None':0, 'Back':-0.1}, 'Vlin'),
            'Turn' : ({'Left':30, 'MLeft':10, 'None':0, 'MRight':-10, 'Right':-30}, 'Vrot')
        }

        self.frules = {
        }

        # Set the degree of achievement
        self.fgoal = "True"

        # Finally, initialize the fuzzy predicats and fuzzy output sets
        self.init_flsets()
        self.init_fpreds()


class Open (Behavior):
    """
    Open a door by a given name, just calling the relevant ROS2 service
    """
    def __init__(self, door):
        print("Instantiated Open behavior with door =", door)
        self.doorname = door
        super().__init__()
    
    def update_state(self, state):

        robot_gwy.execute_door_command(self.doorname,"Open")

    def setup(self):
        self.fgoal = "True"


class Close (Behavior):
    """
    Close a door by a given name, just calling the relevant ROS2 service
    """
    def __init__(self, door):
        self.doorname = door
        print("Instantiated Close behavior with door =", door)
        super().__init__()

    def update_state(self, state):
        robot_gwy.execute_door_command(self.doorname,"Close")

    def setup(self):
        self.fgoal = "True"


class Pick (Behavior):
    """
    Pick a box by a given name, just calling the relevant ROS2 service
    """
    def __init__(self, box):
        self.box = box
        print("Instantiated Pick behavior with box ", box)
        super().__init__()

    def update_state(self, state):
        robot_gwy.pick_up_box(self.box)

    def setup(self):
        self.fgoal = "True"


class Place (Behavior):
    """
    Place a box by a given name, just calling the relevant ROS2 service
    """
    def __init__(self, box):
        self.box = box
        print("Instantiated Place behavior with box ", box)
        super().__init__()

    def update_state(self, state):
        robot_gwy.put_down_box(self.box)

    def setup(self):
        self.fgoal = "True"


class Controller ():
    def __init__(self):
        self.behavior = None        # top-level fuzzy behavior run by the controller
        self.achieved = 0.0         # level of achievement of current behavior
        self.vlin = 0.0             # current value for vlin control variable
        self.vrot = 0.0             # current value for vrot control variable

    def set_behavior (self, bname = None, bparam = None):
        """
        Initialize the controller, setting the behavior to be executed
        Return True if the inizialization is successful
        """
        from robot_map import map
        if bname:
            self.behavior = globals()[bname](bparam)
        return True

    def run (self, state, debug):
        """
        Action decision. Compute control values (vlin, vrot) given the current robot's pose
        by running the current behavior; return the level of achievement of that behavior
        """
        self.achieved = self.behavior.run(state, debug)
        self.vlin = self.behavior.get_vlin()
        self.vrot = self.behavior.get_vrot()
        if debug > 1:
            print('Goal achievement: {:.2f}'.format(self.achieved))
        if debug > 1:
            print('(vlin, vrot)) = ({:.2f}, {:.2f})'.format(self.vlin, degrees(self.vrot))) 
        return self.achieved

    def get_vlin (self):
        return self.vlin
    
    def get_vrot (self):
        return self.vrot
    
