"""Planner based on the Unified Planning library."""

from unified_planning.shortcuts import (
    And,
    Bool,
    BoolType,
    Equals,
    Fluent,
    InstantaneousAction,
    Int,
    MinimizeActionCosts,
    Not,
    Object,
    OneshotPlanner,
    PlanValidator,
    Problem,
    UserType,
    Variable,
    get_environment,
)
from unified_planning.engines import PlanGenerationResultStatus
from unified_planning.io import PDDLWriter

# Disable credits output.
get_environment().credits_stream = None


ORIGIN_SUFFIX = "-origin"


def plan(state, goal):
    print("State:", state)
    print("Goal:", goal)
    task = _get_planning_task(state, goal)
    print(task)
    plan = _solve_planning_task(task)
    return _convert_plan_for_controller(plan)


def _get_planning_task(state, goal):
    # Declare types.
    Room = UserType("Room")
    Locatable = UserType("Locatable")
    Landmark = UserType("Landmark", father=Locatable)
    Door = UserType("Door", father=Landmark)
    Movable = UserType("Movable", father=Locatable)
    Box = UserType("Box", father=Movable)
    Robot = UserType("Robot", father=Movable)

    # Declare predicates.
    in_room = Fluent("in_room", BoolType(), l=Locatable, r=Room)
    at = Fluent("at", BoolType(), m=Movable, l=Landmark)
    # TODO: add predicate for whether a door is closed.
    # TODO: add predicate for whether a door connects two rooms.
    # TODO: add predicate for which object a robot is carrying.

    # Add (typed) objects to problem.
    problem = Problem("household")

    doors = set()
    rooms = set()
    for door, (room1, room2) in state.connects.items():
        doors.add(door)
        rooms.add(room1)
        rooms.add(room2)

    for door in doors:
        problem.add_object(Object(door, Door))

    for room in rooms:
        problem.add_object(Object(room, Room))

    robot = Object("me", Robot)
    openspace = Object("openspace", Landmark)
    problem.add_objects([robot, openspace])

    for entity in state.room:
        if entity.startswith("box") and not entity.endswith(ORIGIN_SUFFIX):
            problem.add_object(Object(entity, Box))
        elif entity != "me":
            problem.add_object(Object(entity, Landmark))

    # Specify the initial state.
    problem.add_fluent(in_room, default_initial_value=False)
    problem.add_fluent(at, default_initial_value=False)
    # TODO: add fluent for whether a door is closed.
    # TODO: add fluent for whether a door connects two rooms.
    # TODO: add fluent for which object a robot is carrying.
    # Each room has an "openspace" landmark.
    for room in problem.objects(typename=Room):
        problem.set_initial_value(in_room(openspace, room), True)
    for entity, room in state.room.items():
        problem.set_initial_value(in_room(problem.object(entity), problem.object(room)), True)
    for entity, landmark in state.pos.items():
        problem.set_initial_value(at(problem.object(entity), problem.object(landmark)), True)
    for door, status in state.door.items():
        # TODO: set initial values for door statuses.
        pass
    for door, (room1, room2) in state.connects.items():
        room1 = problem.object(room1)
        room2 = problem.object(room2)
        door = problem.object(door)
        # TODO: set which doors connect which rooms and which two rooms each door is in.
    if state.carrying:
        # TODO: set initial value for which object the robot is carrying.
        pass

    # Add actions.

    # GoTo(from_landmark, to_landmark, in_room)
    #   preconditions: landmark is in the same room as the robot
    #   effect: move robot near the landmark
    goto = InstantaneousAction("GoTo", from_landmark=Landmark, to_landmark=Landmark, in_room=Room)
    from_landmark = goto.parameter("from_landmark")
    to_landmark = goto.parameter("to_landmark")
    room = goto.parameter("in_room")
    goto.add_precondition(Not(Equals(to_landmark, openspace)))  # robot cannot go to "openspace"
    goto.add_precondition(at(robot, from_landmark))
    goto.add_precondition(in_room(from_landmark, room))
    goto.add_precondition(in_room(to_landmark, room))
    goto.add_precondition(in_room(robot, room))
    goto.add_effect(at(robot, from_landmark), False)
    goto.add_effect(at(robot, to_landmark), True)
    problem.add_action(goto)

    # Open(door)
    #   preconditions: none (it's a radio command to an automatic door)
    #   effect: open the door
    open = InstantaneousAction("Open", door=Door)
    # TODO: add preconditions and effects, then uncomment next line.
    #problem.add_action(open)

    # Close(door)
    #   preconditions: none (it's a radio command to an automatic door)
    #   effect: close the door
    close = InstantaneousAction("Close", door=Door)
    # TODO: add preconditions and effects, then uncomment next line.
    #problem.add_action(close)

    # Cross(from_landmark, from_room, door, to_room)
    #   preconditions: the robot is in front of the door (at a reasonable distance), and the door is open
    #   effect: get the robot to the other side of the door
    cross = InstantaneousAction("Cross", from_landmark=Landmark, from_room=Room, door=Door, to_room=Room)
    # TODO: add preconditions and effects, then uncomment next line.
    #problem.add_action(cross)

    # Pick(box, landmark)
    #   preconditions: the robot and the box are at the same landmark and the robot is not carrying anything
    #   effect: the robot is carrying the box
    pick = InstantaneousAction("Pick", box=Box, landmark=Landmark)
    # TODO: add preconditions and effects, then uncomment next line.
    #problem.add_action(pick)

    # Place(box, landmark)
    #   preconditions: the robot is at the landmark and carrying the box
    #   effect: the box is put down at the landmark and no longer carried by the robot
    place = InstantaneousAction("Place", box=Box, landmark=Landmark)
    # TODO: add preconditions and effects, then uncomment next line.
    #problem.add_action(place)

    # Specify the goal.
    command, *args = goal
    if command == "navigate_to":
        target_location = args[0]
        problem.add_goal(at(robot, problem.object(target_location)))
    elif command == "transport_to":
        box, target_location = args
        problem.add_goal(And(at(problem.object(box), problem.object(target_location))))
    else:
        raise ValueError(f"Unknown command: {command}")

    # We want to minimize the plan cost. Moving within a room is free.
    problem.add_quality_metric(MinimizeActionCosts({cross: 1, pick: 1, place: 1}, default=Int(1)))
    return problem


def _solve_planning_task(problem: Problem):
    # For debugging.
    # w = PDDLWriter(problem)
    # w.write_domain('domain.pddl')
    # w.write_problem('problem.pddl')

    params = {"fast_downward_search_config": "eager_greedy([ff()])"}

    with OneshotPlanner(name="fast-downward", params=params) as p:
        result = p.solve(problem)
        if result.status == PlanGenerationResultStatus.SOLVED_SATISFICING:
            print("Found a plan of length:", len(result.plan.actions))
            print(result.plan)
            with PlanValidator() as validator:
                val_result = validator.validate(problem, result.plan)
                print(val_result.metric_evaluations)
            return result.plan
        else:
            print("No plan found.")
            raise ValueError("No plan found.")


def _convert_plan_for_controller(plan):
    def convert_action(action):
        name = action.action.name
        if name == "GoTo":
            from_landmark, to_landmark, room = action.actual_parameters
            return ("GoTo", to_landmark.object().name.replace(ORIGIN_SUFFIX, ""))
        elif name == "Open":
            door = action.actual_parameters[0]
            return ("Open", door.object().name)
        elif name == "Close":
            door = action.actual_parameters[0]
            return ("Close", door.object().name)
        elif name == "Cross":
            landmark, room1, door, room2 = action.actual_parameters
            return ("Cross", door.object().name)
        elif name == "Pick":
            box, landmark = action.actual_parameters
            return ("Pick", box.object().name)
        elif name == "Place":
            box, landmark = action.actual_parameters
            return ("Place", box.object().name)
        else:
            raise ValueError("Unknown action: " + name)

    return [convert_action(action) for action in plan.actions]


def _test():
    from state import State

    s = State()
    s.pos = {
        "me": "openspace",
        "box1": f"box1{ORIGIN_SUFFIX}",
        "box2": f"box2{ORIGIN_SUFFIX}",
        "box3": f"box3{ORIGIN_SUFFIX}",
    }
    s.room = {
        "bed1": "room1",
        "wardrobe1": "room1",
        "bookshelf1": "room1",
        "bookshelf2": "room1",
        "bookshelf3": "room4",
        "sofa1": "room4",
        "piano1": "room4",
        "fridge1": "room2",
        "sink1": "room2",
        "stove1": "room2",
        "table3": "room2",
        "entrance": "room3",
        "table1": "room4",
        "table2": "room4",
        "me": "room4",
        "box1": "room1",
        f"box1{ORIGIN_SUFFIX}": "room1",
        "box2": "room4",
        f"box2{ORIGIN_SUFFIX}": "room4",
        "box3": "room4",
        f"box3{ORIGIN_SUFFIX}": "room4",
    }
    s.door = {
        "door1": "closed",
        "door2": "closed",
        "door3": "closed",
        "door4": "closed",
    }
    s.connects = {
        "door1": ("room4", "room3"),
        "door2": ("room4", "room1"),
        "door3": ("room3", "room1"),
        "door4": ("room3", "room2"),
    }
    goal = ('navigate_to', 'table1')  # navigate within one room
    #goal = ('navigate_to', 'fridge1')  # navigate between rooms
    #goal = ("transport_to", "box2", "wardrobe1")  # transport box between rooms

    print(plan(s, goal))


if __name__ == "__main__":
    _test()
