"""
============== UniBo: AI and Robotics 2024 ==============
Base code: gateway to the robot (or simulator)
This is a dummy version, most values and funtions are place-holders
It needs to be customized by the students for the different labs

(c) 2024 Alessandro Saffiotti
"""

import rclpy
from geometry_msgs.msg import Twist

from nav_msgs.msg import Odometry
from geometry_msgs.msg import Twist
import rclpy.executors
from sensor_msgs.msg import Range
import math
import numpy as np
import time

from atlantis_msgs.srv import DoorCommand
from atlantis_msgs.srv import DoorStatus
from atlantis_msgs.srv import BoxPos
from atlantis_msgs.srv import BoxPickUp
from atlantis_msgs.srv import BoxPutDown

node = None

robot_subs = None
sonar_subs = []
real_pose = None
vel_publisher_node = None

cmd_vel_publisher = None
doors_clients = None
doors_status = None
boxes_pos_clients = None
boxes_pick_clients = None
boxes_drop_clients = None
sonars = [0.0] * 12

class Quaternion:
    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

def euler_from_quaternion(quaternion):
    x = quaternion.x
    y = quaternion.y
    z = quaternion.z
    w = quaternion.w

    # Normalize quaternion to avoid errors
    norm = np.sqrt(x**2 + y**2 + z**2 + w**2)
    x /= norm
    y /= norm
    z /= norm
    w /= norm

    # Roll
    roll = np.arctan2(2.0 * (w * x + y * z), 1.0 - 2.0 * (x * x + y * y))

    # Pitch
    sin_pitch = 2.0 * (w * y - z * x)
    sin_pitch = np.clip(sin_pitch, -1.0, 1.0) 
    pitch = np.arcsin(sin_pitch)

    # Yaw
    yaw = np.arctan2(2.0 * (w * z + x * y), 1.0 - 2.0 * (y * y + z * z))

    return roll, pitch, yaw

def sonar_ring_pose(bearing):
    rho = 0.260
    phi = math.radians(bearing)
    x = rho * math.cos(phi)
    y = rho * math.sin(phi)
    return (x, y, phi)

parameters = {                        # these are for the Tiago robot
    'wheel_radius' : 0.0985,
    'wheel_axis' : 0.4044,
    'sonar_num' : 12,
    'sonar_maxrange' : 3.0,
    'sonar_delta' : math.radians(25.0),
    'sonar_poses' : [sonar_ring_pose(0.0),
                     sonar_ring_pose(30.0),
                     sonar_ring_pose(60.0),
                     sonar_ring_pose(90.0),
                     sonar_ring_pose(120.0),
                     sonar_ring_pose(150.0),
                     sonar_ring_pose(180.0),
                     sonar_ring_pose(210.0),
                     sonar_ring_pose(240.0),
                     sonar_ring_pose(270.0),
                     sonar_ring_pose(300.0),
                     sonar_ring_pose(330.0)]
}


def sonar_callback(i):
    def cb(msg):
        global sonars
        sonars[i] = msg.range
        #robot_subs.get_logger().info('Range: %f' % (msg.range))
    return cb



def real_pose_callback (msg):
    global real_pose
    real_pose = msg.pose.pose
    #node.get_logger().info('Position: "%f, %f, %f"' % (msg.pose.pose.position.x, msg.pose.pose.position.y, msg.pose.pose.position.z))

def is_ready ():
    return real_pose != None

def get_robot_pose():

    executor.add_node(robot_subs)
    start_time = time.time()

    # Spin for a fixed amount of time
    duration = 0.01
    while rclpy.ok():
        # Check elapsed time and break when it exceeds the duration
        if time.time() - start_time > duration:
            break
        executor.spin_once(timeout_sec=0.01) 
    quat = Quaternion(real_pose.orientation.x,
            real_pose.orientation.y,
            real_pose.orientation.z,
            real_pose.orientation.w)
    roll, pitch, yaw = euler_from_quaternion(quat)
    return (real_pose.position.x, real_pose.position.y, yaw)


def init_robot():
    """
    This should set up the communication channels with the robot (or simulator)
    and perform any initialization needed
    """
    print("Robot initialized")
    global cmd_vel_publisher, doors_clients, boxes_pos_clients, doors_status, boxes_pick_clients, boxes_drop_clients, sonars
    global node, vel_publisher_node
    
    
    node = rclpy.create_node('robot_client_nodes')
    vel_publisher_node = rclpy.create_node('vel_publisher_node')
    cmd_vel_publisher = vel_publisher_node.create_publisher(Twist, '/mobile_base_controller/cmd_vel_unstamped', 1000)
    
    doors_clients = {
            'door1': node.create_client(DoorCommand, 'door1/door_command'),
            'door2': node.create_client(DoorCommand, 'door2/door_command'),
            'door3': node.create_client(DoorCommand, 'door3/door_command'),
            'door4': node.create_client(DoorCommand, 'door4/door_command'),
    }

    doors_status = {
            'door1': node.create_client(DoorStatus, 'door1/get_status'),
            'door2': node.create_client(DoorStatus, 'door2/get_status'),
            'door3': node.create_client(DoorStatus, 'door3/get_status'),
            'door4': node.create_client(DoorStatus, 'door4/get_status'),
    }


    boxes_pos_clients = {
            'BOX1': node.create_client(BoxPos, 'BOX1/get_position'),
            'BOX2': node.create_client(BoxPos, 'BOX2/get_position'),
            'BOX3': node.create_client(BoxPos, 'BOX3/get_position')
    }

    boxes_pick_clients = {
            'BOX1': node.create_client(BoxPickUp, 'BOX1/pick_up'),
            'BOX2': node.create_client(BoxPickUp, 'BOX2/pick_up'),
            'BOX3': node.create_client(BoxPickUp, 'BOX3/pick_up')
    }

    boxes_drop_clients = {
            'BOX1': node.create_client(BoxPutDown, 'BOX1/put_down'),
            'BOX2': node.create_client(BoxPutDown, 'BOX2/put_down'),
            'BOX3': node.create_client(BoxPutDown, 'BOX3/put_down')
    }
    init_subs()



def init_subs():
    global robot_subs, sonar_subs, pose_sub, sonars, executor
    robot_subs = rclpy.create_node('robot_subs_nodes')
    pose_sub = robot_subs.create_subscription(
            Odometry,
            '/ground_truth_odom',
            real_pose_callback,
            10)
    
    for i in range(parameters['sonar_num']):
        sonar_sub = robot_subs.create_subscription(
                Range,
                '/sonar{:02d}_base'.format(i+1),
                sonar_callback(i),
                10)
        sonar_subs.append(sonar_sub)
        pose_sub  # prevent unused variable warning

    executor = rclpy.executors.MultiThreadedExecutor()

def get_sonar_data():
    """
    Get current values from the sonar ring
    Returns an array of readings in the form (sonar pose, range)
    Sonar pose is (x, y, th) in robot's base frame
    This dummy version returns range 0.0 for all sonars
    """
    executor.add_node(robot_subs)
    start_time = time.time()

    # Spin for a fixed amount of time
    duration = 0.01
    while rclpy.ok():
        # Check elapsed time and break when it exceeds the duration
        if time.time() - start_time > duration:
            break
        executor.spin_once(timeout_sec=0.01) 


    res = []
    for i in range(parameters['sonar_num']):
        res.append((parameters['sonar_poses'][i], 0.0))

    res = []
    for i in range(parameters['sonar_num']):
        res.append((parameters['sonar_poses'][i], sonars[i]))
        #node.get_logger().info('Data %f' % (sonars[i]))

    return res



def shutdown_robot ():
    """
    This should perform any finalization needed on the robot,
    and close the communication channels
    """
    print("Robot shut")
    node.destroy_node()
    robot_subs.destroy_node()
    rclpy.shutdown()


def get_wheel_encoders ():
    """
    Get current values of wheel encoders, which indicate the current position
    of each wheel in radiants
    """
    return (0.0, 0.0)


def get_box_position (box):
    """
    Virtual sensor to read the current (x,y) position of a given box in global frame
    In simulation, this uses a ad-hoc service implemented in Gazebo
    In a real physical setup, this could be implemented by a set of cameras in the environments
    """
    node.get_logger().info('Requested Position of %s' % (box))
    req = BoxPos.Request()
    req.request = True
    client = boxes_pos_clients.get(box.upper())
    
    future = client.call_async(req)
    rclpy.spin_until_future_complete(node, future)
    response = future.result()
    node.get_logger().info(
        'Position of %s: (%f, %f, %f, %f)' % (box, response.x_pos, response.y_pos, response.z_pos, response.yaw))
    return (response.x_pos, response.y_pos, response.yaw)


def pick_up_box(box):
    node.get_logger().info('Picking up box %s' % (box))
    req = BoxPickUp.Request()
    req.action = True
    client = boxes_pick_clients.get(box.upper())
    future = client.call_async(req)
    rclpy.spin_until_future_complete(node, future)
    response = future.result()
    if(response.success):
        node.get_logger().info('Picked up %s ' % (box))
    else:
        node.get_logger().error('Failed to pick up %s ' % (box))
    return response.success


def put_down_box(box):
    node.get_logger().info('Putting down box %s' % (box))
    req = BoxPutDown.Request()
    req.action = True
    client = boxes_drop_clients.get(box.upper())
    future = client.call_async(req)
    rclpy.spin_until_future_complete(node, future)
    response = future.result()
    if(response.success):
        node.get_logger().info('Put down %s ' % (box))
    else:
        node.get_logger().error('Failed to put down %s ' % (box))
    return response.success



def get_door_status (door):
    """
    Virtual sensor to get the current door status
    In simulation, this uses a ad-hoc service implemented in Gazebo
    In a real physical setup, this could be implemented by a set of cameras in the environments
    """
    node.get_logger().info('Requested status of %s' % (door))
    req = DoorStatus.Request()
    req.action = True
    client = doors_status.get(door)
    future = client.call_async(req)
    rclpy.spin_until_future_complete(node, future)
    response = future.result()
    node.get_logger().info('%s is %s' % (door, response.status))
    return response.status


def execute_door_command(door, command):
    req = DoorCommand.Request()
    node.get_logger().info('Request to %s %s!' % (command, door))
    req.command = command
    client = doors_clients.get(door)
    future = client.call_async(req)
    rclpy.spin_until_future_complete(node, future)
    response = future.result()

    if (response.done):
        node.get_logger().info('%s %s!' % (command, door))
    else:
        node.get_logger().error('Failed to %s %s!' % (command, door))
    
    
def set_vel_values (vlin, vrot):
    """
    Set new linear and rotational velocities for the robot's base
    vlin is m/sec vrot is rad/sec
    Returns True if successful
    """
    msg = Twist()
    msg.linear.x = vlin
    msg.angular.z = vrot
    cmd_vel_publisher.publish(msg)

    return True
