"""
============== AIoD school on AI and Robotics 2024 ==============
Base code: top level execution loop
To be customized in an incremental way for the different labs

(c) 2024 Alessandro Saffiotti
"""

import time

import rclpy

import robot_gwy, robot_map
from state import State

# Use the first import below to run the students' version, the second one to run the teachers' version.
# import controller
import controller

#import simple_planner as planner  # For testing the controller with hardcoded plans.
import simple_planner as planner  # Base planner version for students.


class Timer:
    """
    Timing functions for synchronizing the control loop
    """

    def __init__(self):
        self.last = time.perf_counter()

    def reset(self):
        self.last = time.perf_counter()

    def elapsed(self):
        """Retun the time elapsed since last sync"""
        return time.perf_counter() - self.last

    def sync(self, cycle):
        """Sleep until elapsed time is cycle"""
        current = time.perf_counter()
        elapsed = current - self.last
        if cycle > elapsed:
            time.sleep(cycle - elapsed)


class TopLevelLoop:
    """
    Top level execution loop
    """

    def __init__(
        self,
        goal=[],  # top level task, passed to the planner
        mypose=None,  # robot's starting pose, if None take it from the map
        tcycle=0.1,  # cycle time, in sec
        debug=0,  # debug level, use to decide what debug info to print
        # (0: none, 1: only outer SPA loop, 2: inner control loop, 3: details)
    ):
        self.goal = goal
        self.tcycle = tcycle
        self.debug = debug
        self.mypose = mypose  # estimated robot's pose
        self.pars = {}  # robot's parameters
        self.ctr = None  # controller instance
        self.clock = None  # timer instance

        rclpy.init()
        self.pars = robot_gwy.init_robot()

    def log(self, *msg):
        if self.debug > 0:
            print(*msg, flush=True)

    def print_pose(self, mypose):
        print(
            "pose = ({:.2f}, {:.2f}, {:.2f})\n".format(mypose[0], mypose[1], mypose[2])
        )

    def run(self, maxsteps=0, goal=None):
        """
        Main entry point: run this instance of the Top Level loop
        debug value can be used to selectively print debug information
        """
        if goal:
            self.goal = goal
        if self.mypose == None:
            self.mypose = robot_map.map.startpose

        self.ctr = controller.Controller()
        self.clock = Timer()
        self.sense_plan_act(self.goal, maxsteps)
        robot_gwy.shutdown_robot()

    def sense_plan_act(self, goal, maxsteps=0):
        """
        The outer "sense plan act" loop
        """
        # only consider a single goal
        goal = goal[0]

        state = self.get_state()  # S = Sense
        self.log(f"Planner called from initial state: {state}")

        plan = planner.plan(state, goal)  # P = Plan
        self.log(f"Planner returned plan: {plan}")
        if plan:
            result = self.execute_plan(plan, maxsteps)  # A = Act
            self.log("Plan execution completed!" if result else "Plan execution failed!")
            return result
        else:
            self.log("No plan found!")
        return None

    def get_state(self):
        """
        The "Sense" part of the SPA loop: fill the state with the current values,
        taken from the map or from the robot's or environment's sensors
        """
        from robot_map import map
        state = State()

        # First set the static part of the state, taken from the map
        # see in the map which objects (doors) connect rooms to one another
        for room1 in map.topology:
            for room2 in map.topology:
                if room1 == room2:
                    continue
                for object1 in map.topology[room1]:
                    for object2 in map.topology[room2]:
                        if object1 == object2:
                            state.connects[object1] = (room1, room2)
        # set their status
        for door in state.connects:
            state.door[door] = map.properties[door]
        # see what is the room of each object (except the doors above)
        for room, contents in map.topology.items():
            for object in contents:
                if object in state.connects:
                    continue
                state.room[object] = room

        # Then set the dynamic part, taken from the robot's and environment's sensors
        self.mypose = robot_gwy.get_robot_pose()
        state.room["me"] = map.find_room(self.mypose)
        state.pos["me"] = map.find_location(self.mypose)
        for box in map.taxonomy['Box']:
            box_pos = robot_gwy.get_box_position(box)
            state.pos[box] = map.find_location(box_pos)
            state.room[box] = map.find_room(box_pos)
            if state.pos[box] == 'openspace':
                landmark = f"{box}{planner.ORIGIN_SUFFIX}"
                state.pos[box] = landmark
                state.room[landmark] = map.find_room(box_pos)
            map.locations[box] = box_pos + (0.3,)

        # We assume the robot is not carrying any box initially.
        state.carrying = None

        return state

    def execute_plan(self, plan, maxsteps):
        """
        The "Act" part of the SPA loop
        Pop each action in the plan in sequence, and execute it
        Return True for successful plan execution, False for failure
        """
        self.log("Executing plan")
        for action in plan:
            result = self.execute_action(action, maxsteps)
            if result == False:
                break
        return result

    def execute_action(self, action, maxsteps = 0, threshold = 0.7):
        """
        The inner action execution loop: execute 'action'
        until its degree of achievement is greter than 'threshold'
        times out with failure after 'maxsteps' steps (zero = no timeout)
        Return True for successful execution, False for failure
        """
        from robot_map import map

        self.log("Executing action:", action)
        nsteps = 0

        # set the behavior to be run by the controller
        # because these call a service using the door's name
        behavior = action[0]
        if behavior == "Open" or behavior == "Close":
            param = action[1]
        elif behavior == "Pick" or behavior == "Place":
            param = action[1]
        else:
            param = map.locations[action[1]]
        self.ctr.set_behavior(bname=behavior, bparam=param)

        # run the main control pipeline until completion, or failure
        while True:
            if not rclpy.ok():  # ROS was killed
                return False
            nsteps += 1
            if maxsteps > 0 and nsteps < maxsteps:  # timeout
                self.log("Max number of steps reached: exiting")
                return False
            result, done = self.step()  # execute control pipeline
            if result == False:  # behavior failure
                self.log("Action {action} failed")
                return False  # action failed
            if done > threshold:  # behavior completed
                self.log("Action {action} completed")
                return True  # action completed

    def step(self):
        """
        The basic control pipeline: read sensors, estimate state, decide controls, send controls
        Return True for successful execution, plus the current degree of achievement
        """

        if not robot_gwy.is_ready():    # wait until topics are available
            return True, 0.0
        # read proprioceptive sensors (wheel rotations)
        wl, wr = robot_gwy.get_wheel_encoders()
        # read exteroceptive sensors (sonars)
        sdata = robot_gwy.get_sonar_data()
        #self.log("sonars =", " ".join(["{:.2f}".format(s[1]) for s in sdata]))
        if self.debug > 1:
            print("sonars =", " ".join(["{:.2f}".format(s[1]) for s in sdata]))
        # Getting the pose from the gwy
        self.mypose = robot_gwy.get_robot_pose()

        state = {
            "mypose": self.mypose,
            "sdata": sdata,
        }  # state passed to the controller
        achieved = self.ctr.run(state, self.debug)  # decide controls (robot's vels)
        vlin = self.ctr.get_vlin()  # retrieve computed controls (vlin)
        vrot = self.ctr.get_vrot()  # retrieve computed controls (vrot)
        robot_gwy.set_vel_values(vlin, vrot)  # send controls
        if self.debug > 1:
            self.print_pose(self.mypose)
        self.clock.sync(self.tcycle)
        if self.debug == 10:
            print("Cycle = {:.3f}".format(self.clock.elapsed()))
        self.clock.reset()
        # time.sleep(self.tcycle)
        return True, achieved
